#ifndef ASM_COMMON_H_
#define ASM_COMMON_H_

#ifdef __cplusplus
extern "C" {
#endif

/* modlimb_invert() sets inv to the multiplicative inverse of n modulo
   2^GMP_LIMB_BITS, ie. satisfying inv*n == 1 mod 2^GMP_LIMB_BITS.
   n must be odd (otherwise such an inverse doesn't exist).

   This is not to be confused with invert_limb(), which is completely
   different.

   The table lookup gives an inverse with the low 8 bits valid, and each
   multiply step doubles the number of bits.  See Jebelean "An algorithm for
   exact division" end of section 4 (reference in gmp.texi).

   Possible enhancement: Could use UHWtype until the last step, if half-size
   multiplies are faster (might help under _LONG_LONG_LIMB).

   Alternative: As noted in Granlund and Montgomery "Division by Invariant
   Integers using Multiplication" (reference in gmp.texi), n itself gives a
   3-bit inverse immediately, and could be used instead of a table lookup.
   A 4-bit inverse can be obtained effectively from xoring bits 1 and 2 into
   bit 3, for instance with (((n + 2) & 4) << 1) ^ n.  */

#define modlimb_invert_table  __gmp_modlimb_invert_table
__GMP_DECLSPEC extern const unsigned char  modlimb_invert_table[128];

#if GMP_LIMB_BITS <= 8
#define modlimb_invert(inv,n)                                   \
  do {                                                          \
    mp_limb_t  __n = (n);                                       \
    mp_limb_t  __inv;                                           \
    ASSERT ((__n & 1) == 1);                                    \
    __inv = modlimb_invert_table[(__n/2) & 0x7F]; /*  8 */      \
    ASSERT ((__inv * __n & GMP_NUMB_MASK) == 1);                \
    (inv) = __inv & GMP_NUMB_MASK;                              \
  } while (0)
#else
#if GMP_LIMB_BITS <= 16
#define modlimb_invert(inv,n)                                   \
  do {                                                          \
    mp_limb_t  __n = (n);                                       \
    mp_limb_t  __inv;                                           \
    ASSERT ((__n & 1) == 1);                                    \
    __inv = modlimb_invert_table[(__n/2) & 0x7F]; /*  8 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 16 */      \
    ASSERT ((__inv * __n & GMP_NUMB_MASK) == 1);                \
    (inv) = __inv & GMP_NUMB_MASK;                              \
  } while (0)
#else
#if GMP_LIMB_BITS <= 32
#define modlimb_invert(inv,n)                                   \
  do {                                                          \
    mp_limb_t  __n = (n);                                       \
    mp_limb_t  __inv;                                           \
    ASSERT ((__n & 1) == 1);                                    \
    __inv = modlimb_invert_table[(__n/2) & 0x7F]; /*  8 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 16 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 32 */      \
    ASSERT ((__inv * __n & GMP_NUMB_MASK) == 1);                \
    (inv) = __inv & GMP_NUMB_MASK;                              \
  } while (0)
#else
#if GMP_LIMB_BITS <= 64
#define modlimb_invert(inv,n)                                   \
  do {                                                          \
    mp_limb_t  __n = (n);                                       \
    mp_limb_t  __inv;                                           \
    ASSERT ((__n & 1) == 1);                                    \
    __inv = modlimb_invert_table[(__n/2) & 0x7F]; /*  8 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 16 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 32 */      \
    __inv = 2 * __inv - __inv * __inv * __n;      /* 64 */      \
    ASSERT ((__inv * __n & GMP_NUMB_MASK) == 1);                \
    (inv) = __inv & GMP_NUMB_MASK;                              \
  } while (0)
#endif /* 64 */
#endif /* 32 */
#endif /* 16 */
#endif /* 8 */


/* Multiplicative inverse of 3, modulo 2^GMP_LIMB_BITS.
   0xAAAAAAAB for 32 bits, 0xAAAAAAAAAAAAAAAB for 64 bits. */
#define MODLIMB_INVERSE_3   ((GMP_NUMB_MAX / 3) * 2 + 1)





#ifdef __cplusplus
}
#endif

#endif	/* ASM_COMMON_H_ */
