/* $Id: field_prime.c 101 2004-09-17 15:28:03Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h>
#include <gmp.h>
#include <assert.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include "manu.h"
#include "gmp_hacks.h"
#include "field_def.h"
#include "field_common.h"

#include "use_field.h"

typedef mp_limb_t * ELT;
typedef const mp_limb_t * cELT;

struct field_aux_data {
	struct {	/* For Tonelli-Shanks. */
		ELT	z;
		ELT	hq;
		int	e;
	} sylow[1];
};

/* Fortunately, 2^(kw) is never a prime number for 32|w and k>0.
 * Therefore we can assume that the same storage space is enough to store
 * elements of k as well as the modulus.
 */

/* First forward some declarations. */
void prime_field_add(ELT, cELT, cELT, const struct field *);
void prime_field_neg(ELT, cELT, const struct field *);
void prime_field_sub(ELT, cELT, cELT, const struct field *);
void prime_field_addto(ELT, cELT, const struct field *);
void prime_field_mul(ELT, cELT, cELT, const struct field *);
void prime_field_addmul(ELT, cELT, cELT, const struct field *);
void prime_field_inv(ELT, cELT, const struct field *);
void prime_field_set_int(ELT, int, const struct field *);
void prime_field_set_mpn(ELT, cELT, mp_size_t, const struct field *);
void prime_field_set_mpz(ELT, const mpz_t p, const struct field *);
int prime_field_is_int(ELT, const int, const struct field *);
void prime_field_reduce(ELT, const struct field *);
void prime_field_conjugate(ELT, cELT, const struct field *);

void prime_field_add(mp_limb_t * a,
		const mp_limb_t * b,
		const mp_limb_t * c,
		const struct field * f)
{
	mp_limb_t * t;
	t=FAST_ALLOC((1+f->size)*sizeof(mp_limb_t));
	t[f->size]=mpn_add_n(t,b,c,f->size);
	prime_field_set_mpn(a,t,1+f->size,f);
	FAST_FREE(t);
}
void prime_field_neg(mp_limb_t * a,
		const mp_limb_t * b,
		const struct field * f)
{
	generic_field_set(a,b,f);
	mpn_sub_n(a,f->modulus,a,f->size);
}

void prime_field_sub(mp_limb_t * a,
		const mp_limb_t * b,
		const mp_limb_t * c,
		const struct field * f)
{
	mp_limb_t * t;
	t=FAST_ALLOC(f->size*sizeof(mp_limb_t));
	prime_field_neg(t,c,f);
	prime_field_add(a,b,t,f);
	FAST_FREE(t);
}

void prime_field_addto(mp_limb_t * a,
		const mp_limb_t * b,
		const struct field * f)
{
	prime_field_add(a,a,b,f);
}

void prime_field_mul(mp_limb_t * a,
		const mp_limb_t * b,
		const mp_limb_t * c,
		const struct field * f)
{
	mp_limb_t * t;
	t=FAST_ALLOC((f->size<<1)*sizeof(mp_limb_t));
	mpn_mul_n(t,b,c,f->size);
	prime_field_set_mpn(a,t,f->size<<1,f);
	FAST_FREE(t);
}

void prime_field_addmul(mp_limb_t * a,
		const mp_limb_t * b,
		const mp_limb_t * c,
		const struct field * f)
{
	mp_limb_t * t;
	t=FAST_ALLOC((f->size<<1)*sizeof(mp_limb_t));
	mpn_mul_n(t,b,c,f->size);
	prime_field_set_mpn(t,t,f->size<<1,f);	/* This is allowed */
	prime_field_add(a,a,t,f);
	FAST_FREE(t);
}

void prime_field_inv(mp_limb_t * a,
		const mp_limb_t * b,
		const struct field * f)
{
	mpz_t az,bz,cz;

	mpz_init(az);
	MPZ_INIT_SET_MPN(bz,b,f->size);
	MPZ_INIT_SET_MPN(cz,f->modulus,f->size);

	if (mpz_invert(az,bz,cz) == 0) {
		mpz_gcd(az,bz,cz);
		mpz_out_str(stdout,10,az);
		printf(" is a factor of the modulus\n");
		exit(191);
	}
	mpz_clear(bz);
	mpz_clear(cz);
	prime_field_set_mpz(a,az,f);
	mpz_clear(az);
}

static void prime_field_compute_sylow(const struct field * f)
{
	int e;
	mpz_t p,q,z,r,s,m1;
	gmp_randstate_t	prng;

	if (f->aux_data->sylow->e)
		return;

	/* First get e and q right */

	MPZ_INIT_SET_MPN(p,f->modulus,f->size);
	mpz_init_set(q,p);
	mpz_sub_ui(q,q,1);

	for(e=0;q->_mp_size && (q->_mp_d[0] & 1) == 0;e++) {
		mpz_fdiv_q_2exp(q,q,1);
	}

	/* Now look for a convenient generator */

	mpz_init(z);
	mpz_init_set_ui(r,0);
	mpz_init(s);
	mpz_ui_pow_ui(s,2,e-1);

	mpz_init_set(m1,p);
	mpz_sub_ui(m1,m1,1);

	gmp_randinit_default(prng);

	for(;mpz_cmp(r,m1)!=0;) {
		mpz_urandomm(z,prng,p);
		mpz_powm(z,z,q,p);
		mpz_powm(r,z,s,p);
	}

	mpz_clear(m1);

	/* It's actually (q-1)/2 that we mean to store... */
	mpz_sub_ui(q,q,1);
	mpz_fdiv_q_2exp(q,q,1);

	/* NOW set things in. */
	f->aux_data->sylow->e=e;
	f->aux_data->sylow->z = malloc(f->size * sizeof(mp_limb_t));
	f->aux_data->sylow->hq= malloc(f->size * sizeof(mp_limb_t));
	f_set_mpz(f->aux_data->sylow->z,z);
	f_set_mpz(f->aux_data->sylow->hq,q);
	
	/* big cleanup */
	mpz_clear(s);
	mpz_clear(z);
	mpz_clear(r);
	mpz_clear(q);
	mpz_clear(p);
	gmp_randclear(prng);
}


#if 0
int prime_field_sqrt_mpzs(mp_limb_t * rx,
		const mp_limb_t * sa,
		const struct field * f)
{
	mpz_t p,a,y,x,b,t;
	int r,m;
	
	prime_field_compute_sylow(f);

	MPZ_INIT_SET_MPN(p,f->modulus,f->size);
	MPZ_INIT_SET_MPN(a,sa,f->size);
	mpz_init_set(y,f->aux_data->sylow->z);
	mpz_init(x);
	mpz_init(b);
	mpz_init(t);
	r=f->aux_data->sylow->e;

	mpz_powm(x,a,f->aux_data->sylow->hq,p);

	mpz_mul(b,x,x);mpz_mod(b,b,p);
	mpz_mul(b,b,a);mpz_mod(b,b,p);

	mpz_mul(x,x,a);

	for(;;) {
		mpz_set(t,b);
		for(m=0;mpz_cmp_ui(t,1)!=0;m++) {
			mpz_mul(t,t,t);
			mpz_mod(t,t,p);
		}
		ASSERT(m<=r);
		if (m==0 || m==r) break;

		mpz_ui_pow_ui(t,2,r-m-1);
		mpz_powm(t,y,t,p);
		mpz_mul(y,t,t);
		mpz_mod(y,y,p);
		r=m;
		mpz_mul(x,x,t);
		mpz_mod(x,x,p);
		mpz_mul(b,b,y);
		mpz_mod(b,b,p);
	}

	MPN_SET_MPZ(rx,f->size,x);

	mpz_clear(t);
	mpz_clear(p);
	mpz_clear(y);
	mpz_clear(x);
	mpz_clear(b);
	mpz_clear(a);

	return m==0;
}
#endif

/* Do Tonelli-Shanks algorithm to compute the square root of a mod p */
/* return 1 if a square root could be computed, 0 otherwise */
int prime_field_sqrt(mp_limb_t * x,
		const mp_limb_t * a,
		const struct field * f)
{
	ELT	y,b,t,s;
	mp_size_t ss;
	int	r,m;
	
	f_set(x,a);

	if (f_is_zero(x)) {
		return 1;
	}
	prime_field_compute_sylow(f);

	y = malloc(f->size*sizeof(mp_limb_t));
	b = malloc(f->size*sizeof(mp_limb_t));
	t = malloc(f->size*sizeof(mp_limb_t));

	f_set(y,f->aux_data->sylow->z);
	r=f->aux_data->sylow->e;

	ss = (r-1) / GMP_LIMB_BITS + 1;
	s = malloc(ss*sizeof(mp_limb_t));
	memset(s,0,ss*sizeof(mp_limb_t));
	s[ss-1] = 1UL << ((r-1)%GMP_LIMB_BITS);

	f_exp(x,a,f->aux_data->sylow->hq,f->size);
	f_mul(b,x,x);
	f_mul(b,b,a);
	f_mul(x,x,a);

	for(;;) {
		/* s = 2^(r-1) */
		f_set(t,b);
		for(m=0;!f_is_one(t);m++) {
			f_mul(t,t,t);
		}

		ASSERT(m<=r);
		if (m==0 || m==r) break;

		ss = (r-m-1) / GMP_LIMB_BITS + 1;
		memset(s,0,ss*sizeof(mp_limb_t));
		s[ss-1] = 1UL << ((r-m-1)%GMP_LIMB_BITS);

		f_exp(t,y,s,ss);
		f_mul(y,t,t);

		f_mul(x,x,t);
		f_mul(b,b,y);

		r=m;
	}

	free(y);
	free(b);
	free(t);
	free(s);

	return m==0;
}

void prime_field_set_int(mp_limb_t * a,
		int x,
		const struct field * f)
{
	if (x>=0) {
		memset(a+1,0,(f->size-1)*sizeof(mp_limb_t));
		a[0]=x;
#if 0
		/* Hmmm, who cares ? Didn't I say _lazy_ reduction ? */
		if (f->size==1 && x >= f->modulus[0]) {
			a[0]=x % f->modulus[0];
		}
#endif
	} else {	/* then x<=0 */
		x=-x;
		if (f->size==1 && x >= f->modulus[0]) {
			x=x % f->modulus[0];
		}
		a[0]=x;
		prime_field_sub(a,f->modulus,a,f);
	}
}

void prime_field_set_mpn(mp_limb_t * a,
		const mp_limb_t * p,
		mp_size_t n,
		const struct field * f)
{
	if (n < f->size) {
		memcpy(a,p,n*sizeof(mp_limb_t));
		memset(a+n,0,(f->size-n)*sizeof(mp_limb_t));
	} else {
		/* XXX Super-Watch out ! p must have enough size to store
		 * __n+1__ limbs... This is a requirement due to tdiv_qr
		 * and the fact that the modulus needs not have highest
		 * bit set.
		 */
		mp_limb_t * tmp;
		tmp = FAST_ALLOC((n+1-f->size)*sizeof(mp_limb_t));
		mpn_tdiv_qr(tmp, a,0, p,n, f->modulus,f->size);
		FAST_FREE(tmp);
	}
}

void prime_field_set_mpz(mp_limb_t * a,
		const mpz_t p,
		const struct field * f)
{
	if (p[0]._mp_size==0)
		prime_field_set_int(a,0,f);
	else {
		prime_field_set_mpn(a,p[0]._mp_d,abs(p[0]._mp_size),f);
	}
}

int prime_field_is_int(mp_limb_t * a,
		int x,
		const struct field * f)
{
	int i;
	f->reduce(a,f->base_field);
	for(i=f->size;i<f->size;i++)
		if (a[i]!=0UL) return 0;
	if (x>=0) {
		for(i=1;i<f->size;i++)
			if (a[i]!=0UL) return 0;
		if (f->size==1 && x >= f->modulus[0]) {
			x=x % f->modulus[0];
		}
		return (a[0]==x);
	} else {
		mp_limb_t * t;
		int res;
		x=-x;
		if (f->size==1 && x >= f->modulus[0]) {
			x=x % f->modulus[0];
		}
		t=FAST_ALLOC(f->size * sizeof(mp_limb_t));
		t[0]=x;
		mpn_add_n(t,t,a,f->size);
		res=(memcmp(t,f->modulus,f->size)==0);
		FAST_FREE(t);
		return res;
	}
}

void prime_field_reduce(mp_limb_t * a, const struct field * f)
{
	mp_limb_t quot;
	mpn_tdiv_qr(&quot,a,0,a,f->size,f->modulus,f->size);
}

void prime_field_conjugate(mp_limb_t * a,
		const mp_limb_t * b,
		const struct field * f)
{
	generic_field_set(a,b,f);
}

void prime_field_print(const mp_limb_t * x, const struct field * f)
{
	mpz_t a,p;
	mpz_init(a);
	mpz_init(p);
	MPZ_SET_MPN(a,x,f->size);
	MPZ_SET_MPN(p,f->modulus,f->size);
	mpz_mod(a,a,p);
	mpz_out_str(stdout,10,a);
	mpz_clear(a);
	mpz_clear(p);
}

void prime_field_destroy(struct field * f)
{
	if (f->aux_data->sylow->z) {
		free(f->aux_data->sylow->z);
		free(f->aux_data->sylow->hq);
	}
	free(f->aux_data);
	f->aux_data=NULL;
	generic_field_destroy(f);
}

struct field * new_prime_field(const mp_limb_t * p, const mp_size_t n)
{
	struct field * res;

	res=malloc(sizeof(struct field));
	res->size=n;
	res->modulus=malloc(n*sizeof(mp_limb_t));
	memcpy(res->modulus,p,n*sizeof(mp_limb_t));
	res->degree=1;
	res->base_field	= res;
	res->add	= prime_field_add;
	res->sub	= prime_field_sub;
	res->neg	= prime_field_neg;
	res->addto	= prime_field_addto;
	res->mul	= prime_field_mul;
	res->addmul	= prime_field_addmul;
	res->inv	= prime_field_inv;
	res->set_int	= prime_field_set_int;
	res->set_mpn	= prime_field_set_mpn;
	res->set_mpz	= prime_field_set_mpz;
	res->is_int	= prime_field_is_int;
	res->exp	= generic_field_exp;
	res->eq		= generic_field_eq;
	res->set	= generic_field_set;
	res->reduce	= prime_field_reduce;
	res->conjugate	= prime_field_conjugate;
	res->print	= prime_field_print;
	res->destroy	= prime_field_destroy;
	res->minpoly	= NULL;		/* Does not apply */
	res->sqrt	= prime_field_sqrt;

	res->aux_data	= malloc(sizeof(struct field_aux_data));
	memset(res->aux_data,0,sizeof(struct field_aux_data));

	return res;
}

