/* $Id: z2.c 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

/* This is a standard mploc<z2> module */

/* -- z2.c --
 *
 * This interface is for 2-adics, stored as mpz_t's
 *
 * Actually we tend to twiddle more with the mpn layer here, for obvious
 * reasons. Some helper functions that are also used by z2x have been
 * moved into z2_limbs.c
 *
 *
 * Design choice : always allocate for the full precision. High zeroes
 * are always explicit.
 */

/***************************************************************************/
/* HEADERS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* This is all the GMP stuff I need... */
#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "mploc.h"
#include "z2_limbs.h"
#include "z2.h"
#include "manu.h"
#include "asm/asm.h"
#include "gmp_hacks.h"

/* ATTENTION : No underlying finite field implementation is called. */

struct mplf_private {
	/* NOTHING FOR NOW... */
};

/***************************************************************************/
/* ENTRY POINTS -- FIELD INTERFACE */

/* 
 * mplf_z2_create -- creates the 2-adic integer ring
 *
 * INPUT: none
 * OUTPUT: The struct containing useful information
 */

mplf_t mplf_z2_create(void)
{
	struct mplf_struct * f;
	
	f=malloc(sizeof(struct mplf_struct));
#ifdef DEBUG_MPLOC
	f->people=0;
#endif
	//f->force_precision=&mploc_z2_force_precision;
	// f->priv=malloc(sizeof(struct mplf_private));
	f->degree=1;
	f->ramified=0;
	mpz_init_set_ui(f->p,2);
	return f;
}

/*
 * mplf_z2_destroy -- clears any data associated with a field.
 *
 * INPUT: The field info
 * OUTPUT: none
 * SIDE EFFECT: everything cleaned up ; *pf is set to NULL
 */

void mplf_z2_destroy(mplf_t * pf)
{
#ifdef DEBUG_MPLOC
	ASSERT(!(*pf)->people);
#endif
	// free((*pf)->priv);
	mpz_clear((*pf)->p);
	free((*pf));
	(*pf)=NULL;
}

/***************************************************************************/
/* ENTRY POINTS -- ELEMENT INTERFACE */

/*
 * mploc_z2_force_precision -- change the significance of a padic int.
 *
 * INPUT: the padic number and the new precision
 * OUTPUT: none
 * SIDE EFFECT: x->prec changed. Inner representation updated.
 *
 * COMMENTS: - this has no effect on the stored size when it boils down
 * to less than a wordsize change
 */


#if 0
static void realloc_on_mpz_t_below(mploc_ptr x)
{
	mp_size_t req;
	req=iceildiv(x->prec,GMP_LIMB_BITS);
	if (ALLOC(x->x) < req) {
		PTR(x->x)=realloc(PTR(x->x),req*sizeof(mp_limb_t));
		/* 
		memset(PTR(x->x)+abs(SIZ(x->x)),
				0,abs(req-SIZ(x->x))*sizeof(mp_limb_t));
				*/
	}
}
#endif

/* Performs some consistency tests to make sure nothing's screwed up */
int mploc_z2_is_good_looking(mploc_srcptr x)
{
	if (SIZ(x->x)<0) {
		fprintf(stderr, "mploc_t represented by negative integer\n");
		return 0;
	}

	if (ALLOC(x->x) < SIZ(x->x)) {
		fprintf(stderr, "mploc_t with insufficient allocation\n");
		return 0;
	}

#if 0	/* Enable only to diagnose missing reduction problems */
	if (ALLOC(x->x) > MAX(4*SIZ(x->x),SIZ(x->x)+128)) {
		fprintf(stderr, "mploc_t with unexpectedly high allocation\n");
		return 0;
	}
#endif
	
	if (SIZ(x->x)!=iceildiv(x->prec,GMP_LIMB_BITS)) {
		fprintf(stderr, "mploc_t with incoherent precision\n");
		return 0;
	}

	return 1;
}

void mploc_z2_force_precision(mploc_ptr x, mp_size_t k)
{
	mp_size_t req;

	req=iceildiv(k,GMP_LIMB_BITS);

	if (ALLOC(x->x) < req) {
		mpz_realloc2(x->x,k);
		ALLOC(x->x) = req;
	}

	SIZ(x->x) = req;
	x->prec=k;
	
	ASSERT(mploc_z2_is_good_looking(x));
}

/* Zero out all bits above k ; k must be <= x->prec -- Use this after
 * mploc_z2_force_precision if you want to please purify
 */
void mploc_z2_insert_zero(mploc_ptr x, mp_size_t k)
{
	mploc_limbs_insert_zero(PTR(x->x),SIZ(x->x),k);

	ASSERT(mploc_z2_is_good_looking(x));
}

/*
 * mploc_init -- create an element
 */
void mploc_z2_init(mploc_ptr x, mplf_t f)
{
	mpz_init(x->x);
	x->prec=0;
	x->f=f;
#ifdef DEBUG_MPLOC
	x->f->people++;
#endif
	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_set_change_precision(mploc_ptr x, mploc_srcptr y, mp_size_t k)
{
	ASSERT(mploc_z2_is_good_looking(y));

	mploc_z2_force_precision(x,k);

	if (x != y) {
		mploc_limbs_copy(PTR(x->x),PTR(y->x),MIN(SIZ(x->x),SIZ(y->x)));
	}

	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_set(mploc_ptr x, mploc_srcptr y)
{
	mploc_z2_set_change_precision(x,y,y->prec);
}


void mploc_z2_set_ui(mploc_ptr x, mp_limb_t u, unsigned int prec)
{
	mploc_z2_force_precision(x,prec);
	mploc_limbs_zero(PTR(x->x),SIZ(x->x));
	PTR(x->x)[0]=u;
	
	ASSERT(mploc_z2_is_good_looking(x));
}


void mploc_z2_set_mpz(mploc_ptr x, mpz_srcptr a, unsigned int prec)
{
	mploc_z2_force_precision(x,prec);
	if (mpz_sgn(a) >= 0) {
		if (SIZ(a)<SIZ(x->x)) {
			mploc_limbs_copy(PTR(x->x),PTR(a),SIZ(a));
			mploc_limbs_zero(PTR(x->x)+SIZ(a),SIZ(x->x)-SIZ(a));
		} else {
			mploc_limbs_copy(PTR(x->x),PTR(a),SIZ(x->x));
		}
	} else {
		if (-SIZ(a)<SIZ(x->x)) {
			mploc_limbs_copy(PTR(x->x),PTR(a),-SIZ(a));
			mploc_limbs_zero(PTR(x->x)-SIZ(a),SIZ(x->x)+SIZ(a));
		} else {
			mploc_limbs_copy(PTR(x->x),PTR(a),SIZ(x->x));
		}
		mploc_z2_neg(x, x);
	}

	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_set_random(mploc_ptr x, gmp_randstate_t prng, unsigned int prec)
{
	/* Unfortunately no mpn interface is provided for mpz_urandomb --
	 * We'll duplicate things at the z2x level */
	mpz_urandomb(x->x,prng,prec);
	mploc_z2_force_precision(x,prec);

	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_add(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	mp_size_t nw;

	ASSERT(mploc_z2_is_good_looking(x));
	ASSERT(mploc_z2_is_good_looking(y));

	if (EXPECT_FALSE(x->prec==0 || y->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	z->prec = MIN(x->prec,y->prec);
	nw	= MIN(SIZ(x->x),SIZ(y->x));
	if (ALLOC(z->x) < nw) {
		PTR(z->x)=realloc(PTR(z->x),nw*sizeof(mp_limb_t));
		ALLOC(z->x) = nw;
	}
	SIZ(z->x) = nw;
	mpn_add_n(PTR(z->x),PTR(y->x),PTR(x->x),nw);

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_sub(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	mp_size_t nw;

	ASSERT(mploc_z2_is_good_looking(x));
	ASSERT(mploc_z2_is_good_looking(y));

	if (EXPECT_FALSE(x->prec==0 || y->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	z->prec = MIN(x->prec,y->prec);
	nw	= MIN(SIZ(x->x),SIZ(y->x));
	if (ALLOC(z->x) < nw) {
		PTR(z->x)=realloc(PTR(z->x),nw*sizeof(mp_limb_t));
		ALLOC(z->x) = nw;
	}
	SIZ(z->x) = nw;
	mpn_sub_n(PTR(z->x),PTR(x->x),PTR(y->x),nw);

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_normalize(mploc_ptr z UNUSED_VARIABLE)
{
	/* We're always normalized, apparently */
}

void mploc_z2_neg(mploc_ptr z, mploc_srcptr x)
{
	ASSERT(mploc_z2_is_good_looking(x));

	mploc_z2_force_precision(z,x->prec);

	mploc_limbs_neg(PTR(z->x),PTR(x->x),SIZ(z->x));

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_add_ui(mploc_ptr z, mploc_srcptr x, mp_limb_t u)
{
	ASSERT(mploc_z2_is_good_looking(x));

	if (EXPECT_FALSE(x->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	mploc_z2_force_precision(z,x->prec);
	mpn_add_1(PTR(z->x),PTR(x->x),SIZ(x->x),u);

	ASSERT(mploc_z2_is_good_looking(z));

}
void mploc_z2_sub_ui(mploc_ptr z, mploc_srcptr x, mp_limb_t u)
{
	ASSERT(mploc_z2_is_good_looking(x));

	if (EXPECT_FALSE(x->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	mploc_z2_force_precision(z,x->prec);
	mpn_sub_1(PTR(z->x),PTR(x->x),SIZ(x->x),u);

	ASSERT(mploc_z2_is_good_looking(z));
}

/* Ayee ! The precision *increases* */
void mploc_z2_sqr(mploc_ptr z, mploc_srcptr x)
{
	mp_size_t	nw;
	mp_limb_t	c, * t;
	mp_size_t	xprec;

	ASSERT(mploc_z2_is_good_looking(x));

	if (EXPECT_FALSE(x->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	/*
	 * There's a subtle issue here, since the precision increases,
	 * and we want to give this back to the user, since he wants it
	 * (he precisely called _sqr)
	 *
	 * We can't simply increase the precision here in z since we would
	 * easily obtain a segfault (x needs not have that much stuff
	 * allocated, and we do touch it !!)
	 *
	 * So we do otherwise ; see carry mechanism in
	 * mploc_limbs_short_square
	 */

	nw	= SIZ(x->x);
	/* 20040824 ; ah, good one ;-) */
	xprec=x->prec;
	mploc_z2_force_precision(z,x->prec+1);

	t = FAST_ALLOC(2 * nw * sizeof(mp_limb_t));

	/* 20030223: weird, the ASSERT down below within the if statement
	 * was here, which is obviously incorrect.
	 */
	ASSERT(z==x || nw==SIZ(x->x));

	if ((xprec & (GMP_LIMB_BITS-1)) == 0) {
		ASSERT(SIZ(z->x)==nw + 1);
		c = mploc_limbs_short_square_carry(PTR(z->x),PTR(x->x),t,nw);
		PTR(z->x)[SIZ(z->x)-1]=c;
	} else {
		mploc_limbs_short_square(PTR(z->x),PTR(x->x),t,nw);
	}
	FAST_FREE(t);

	/* What did we get on return ? Bit of order
	 * far = SIZ(z->x)*GMP_LIMB_BITS. Two cases.
	 *
	 * far = z->prec ; Then do mploc_z2_force_precision on z, since
	 * more space needs to be allocated, and then set the new limb
	 * accordingly.
	 *
	 * far > z->prec ; Then c is meaningless. We can safely do
	 * z->prec++, and the corresponding bit will have been correctly
	 * handled by mploc_limbs_short_square.
	 */

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_mul(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	mp_size_t	nw;
	mp_limb_t	* t;

	ASSERT(mploc_z2_is_good_looking(x));
	ASSERT(mploc_z2_is_good_looking(y));

	if (EXPECT_FALSE(x->prec==0 || y->prec==0)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	if (EXPECT(x==y,0)) {
		mploc_z2_sqr(z,x);
		return;
	}

	mploc_z2_force_precision(z,MIN(x->prec,y->prec));
	nw	= SIZ(z->x);

	t = FAST_ALLOC(2*nw*sizeof(mp_limb_t));

	if (z == x || z == y) {
		mp_limb_t * d;
		d = malloc(MAX(ALLOC(z->x),nw) * sizeof(mp_limb_t));
		mploc_limbs_short_product(d,PTR(x->x),PTR(y->x),t,nw);
		free(PTR(z->x));
		PTR(z->x) = d;
		ALLOC(z->x)=MAX(ALLOC(z->x),nw);
	} else {
		mploc_limbs_short_product(PTR(z->x),PTR(x->x),PTR(y->x),t,nw);
	}
	FAST_FREE(t);

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_rshift(mploc_ptr z, mploc_srcptr x, mp_size_t n)
{
	mp_size_t ws,ss;

	ASSERT(mploc_z2_is_good_looking(x));

	if (EXPECT_FALSE(n==0)) {
		mploc_z2_set(z,x);
	}

	if (EXPECT_FALSE(x->prec <= n)) {
		mploc_z2_force_precision(z,0);
		return;
	}

	/* start with x->prec in case z==x */
	mploc_z2_force_precision(z,x->prec);

	ws = n / GMP_LIMB_BITS;
	ss = n % GMP_LIMB_BITS;

	ASSERT(ws < SIZ(x->x)); /* otherwise we would have had x->prec <= n */

	if (EXPECT_FALSE(ss==0)) {
		mploc_limbs_move(PTR(z->x),PTR(x->x)+ws,SIZ(x->x)-ws);
	} else {
		/* Note that the gmp manual specifies that overlapping is
		 * ok here, since the result ptr is before the source.
		 * (it is counter-intuitive indeed).
		 */
		mpn_rshift(PTR(z->x),PTR(x->x)+ws,SIZ(x->x)-ws,ss);
	}

	/* This is a bit crude... */
	z->prec-=n;
	SIZ(z->x) = iceildiv(z->prec,GMP_LIMB_BITS);

	ASSERT(mploc_z2_is_good_looking(z));
}

void mploc_z2_lshift(mploc_ptr z, mploc_srcptr x, mp_size_t n)
{
	mp_size_t xcw,wi,wp;
	int needpop;

	/* This code has been backported from mploc_z2x_lshift */

	ASSERT(mploc_z2_is_good_looking(x));

	if (EXPECT_FALSE(n==0)) {
		mploc_z2_set(z,x);
		return;
	} else if (EXPECT_FALSE(x->prec==0)) {
		mploc_z2_force_precision(z,x->prec+n);
		/* This is mandatory */
		mploc_z2_insert_zero(z,0);
		return;
	}

	/* Save it now so that self-assignment doesn't kill us */
	xcw = SIZ(x->x);

	wi = n / GMP_LIMB_BITS;
	wp = n % GMP_LIMB_BITS;

	needpop = (((x->prec-1) % GMP_LIMB_BITS)+1 + wp) > GMP_LIMB_BITS;

	mploc_z2_force_precision(z,x->prec+n);

	ASSERT(SIZ(z->x)-wi >= xcw && ((SIZ(z->x)-wi > xcw) == needpop));

	if (EXPECT_FALSE(wp==0)) {
		mploc_limbs_move(PTR(z->x)+wi,PTR(x->x),xcw);
	} else if (needpop) {
		mp_limb_t pop;
		pop = mpn_lshift(PTR(z->x)+wi,PTR(x->x),xcw,wp);
		PTR(z->x)[wi+xcw]=pop;
	} else {
		/* TODO : maybe merge with branch above ? */
		BAD_CODE_WARNING;
		mpn_lshift(PTR(z->x)+wi,PTR(x->x),xcw,wp);
	}
	mploc_limbs_zero(PTR(z->x),wi);
	/* caveat here ; zeroing must happen _after_ the shift, of
	 * course !
	 */

	ASSERT(mploc_z2_is_good_looking(z));
}

/* slooooow */
void mploc_z2_print(mploc_srcptr x)
{
	mpz_t q,r;
	int i;
	
	ASSERT(mploc_z2_is_good_looking(x));

	mpz_init(q);
	mpz_init(r);
	mpz_set(q,x->x);
	MPN_NORMALIZE(PTR(q),SIZ(q));
	mploc_limbs_insert_zero(PTR(q),SIZ(q),x->prec);
	for(i=0;i<x->prec && (mpz_cmp_ui(q,0)!=0 || (i>x->prec-4)); i++) {
		mpz_fdiv_qr_ui(q,r,q,2);
		gmp_printf("%Zd",r);
	}
	if (i<x->prec)
		printf("0<%ld>",x->prec-i);
	printf("...");
	mpz_clear(q);
	mpz_clear(r);
	printf(" + O(2^%d)", (int) x->prec);
}

/*
 * mploc_z2_valuation -- compute the valuation
 *
 * This function is expected to return 0 in most cases. Therefore it is
 * not meant to be efficient for the cases where an element has a big
 * valuation.
 *
 * If x->prec or more is returned, of course one shall assume that the
 * result is not significant (it can be anything between x->prec and
 * +infty)
 */

int mploc_z2_valuation(mploc_srcptr x)
{
	ASSERT(mploc_z2_is_good_looking(x));
	return MIN(mploc_limbs_valuation(PTR(x->x),SIZ(x->x)),x->prec);
}

/*
 * mploc_z2_inv -- compute an inverse.
 *
 * We find by Newton iteration a zero of f(x)=ax-1
 *
 * iteration is x <- x - x(a*x-1)
 *                <- 2*x - a*x^2
 *
 * The code in here can be largely improved
 */

void mploc_z2_inv_improve_iterative(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t	t;

	ASSERT(mploc_z2_is_good_looking(a));

	mploc_z2_init(t,a->f);
	mploc_z2_force_precision(t,2 * a->prec);

	for(;x->prec<aim;) {
		mp_size_t np = 2 * x->prec;
		mp_size_t osz;

		ASSERT(mploc_z2_is_good_looking(t));

		ASSERT(ALLOC(x->x) * GMP_LIMB_BITS >= np);
		ASSERT(ALLOC(t->x) * GMP_LIMB_BITS >= np);
		
		mploc_z2_force_precision(t,np);
		mpn_mul_n(PTR(t->x),PTR(x->x),PTR(x->x),SIZ(x->x));
		mploc_z2_mul(t,t,a);

		osz = SIZ(x->x);

		mploc_z2_force_precision(x,np);
		mploc_z2_insert_zero(x,osz*GMP_LIMB_BITS);
		mploc_z2_lshift(x,x,1);	/* Increases precision ! */

		mploc_z2_sub(x,x,t);
	}

	mploc_z2_clear(t);

	ASSERT(mploc_z2_is_good_looking(x));
}


void mploc_z2_inv_improve_recursive(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t t;

	BAD_CODE_WARNING;	/* Untested, actually */

	ASSERT(mploc_z2_is_good_looking(a));

	if (x->prec >= aim) return;

	mploc_z2_init(t,a->f);
	mploc_z2_init(y,a->f);

	mploc_z2_inv_improve_recursive(x,a,aim - aim/2);
	
	mploc_z2_set_change_precision(t,a,aim);
	mpz_mul(y->x,t->x,x->x);	/* This is a FULL product */
	mploc_z2_sub_ui(y,y,1);
	mpz_mul(y->x,y->x,x->x);
	mploc_z2_neg(y,y);
	mploc_z2_set_change_precision(t,y,aim);
	mploc_z2_add(x,x,t);
	x->prec = aim;

	mploc_z2_clear(t);
	mploc_z2_clear(y);

	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_inv_improve_recursive2(mploc_ptr x, mploc_srcptr a,
		mp_limb_t * buf,
		mp_size_t n)
{
	mp_size_t p;

	ASSERT(mploc_z2_is_good_looking(a));

	ASSERT(ALLOC(x->x)>=2*SIZ(a->x));
	ASSERT(SIZ(a->x)>=n);

	if (SIZ(x->x) >= n) return;

	p=(n+1)>>1;

	mploc_z2_inv_improve_recursive2(x,a,buf,p);
	ASSERT(SIZ(x->x)==p);

	mpn_mul_n(buf+n,PTR(x->x),PTR(x->x),p);
	mploc_limbs_short_product(buf,PTR(a->x),buf+n,buf+2*n,n);
	//memset(x+p,0,n-p);
	//mpn_lshift(PTR(x->x),PTR(x->x),n,1);
	//mpn_sub_n(PTR(x->x),PTR(x->x),buf,n);
	mploc_limbs_neg(PTR(x->x)+p,buf+p,n-p);
	SIZ(x->x)=n;
}

void mploc_z2_inv_improve_recursive2_wrap(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mp_limb_t * buf;

	buf = malloc(4*SIZ(a->x)*sizeof(mp_limb_t));

	mploc_z2_inv_improve_recursive2(x,a,buf,SIZ(a->x));

	x->prec = a->prec;

	free(buf);
}

static int inv_method=2;
static int inv_seed=0;
void mploc_z2_inversion_configure(int new_method, int new_seed)
{
	inv_method=new_method;
	inv_seed=new_seed;
}

void mploc_z2_inv_improve(mploc_ptr x, mploc_srcptr a)
{
	ASSERT(mploc_z2_is_good_looking(a));

	switch(inv_method) {
		case 0: mploc_z2_inv_improve_iterative(x,a,a->prec);break;
		case 1: mploc_z2_inv_improve_recursive(x,a,a->prec);break;
		case 2: mploc_z2_inv_improve_recursive2_wrap(x,a,a->prec);break;
	}

	ASSERT(mploc_z2_is_good_looking(x));
}

void mploc_z2_inv(mploc_ptr z, mploc_srcptr a)
{
	mploc_t x;
	mp_limb_t w;
#if 0
	mp_limb_t a0;
#endif

	ASSERT(mploc_z2_is_good_looking(a));

	if (EXPECT_FALSE(mploc_z2_valuation(a)!=0)) {
		DIVIDE_BY_ZERO();
		return;
	}
		
#if 0

	/* Compute the first word */
	w = a0 = PTR(a->x)[0];
	if (a->prec <= 3) goto done_word;
	w = w ^ ((w&2)<<2) ^ ((w&4)<<1);
	if (a->prec <= 4) goto done_word;
	w = w - w * ((a0*w)&0xf0UL);
	if (a->prec <= 8) goto done_word;
	w = w - w * ((a0*w)&0xff00UL);
#if GMP_LIMB_BITS > 16
	if (a->prec <= 16) goto done_word;
	w = w - w * ((a0*w)&0xffff0000UL);
#if GMP_LIMB_BITS > 32
	if (a->prec <= 32) goto done_word;
	w = w - w * ((a0*w)&0xffffffff00000000UL);
#if GMP_LIMB_BITS > 64
	BUG();	/* Fix it yourself ! */
#endif
#endif
#endif
#else
	/* Exists in GMP !!! */
	modlimb_invert(w,PTR(a->x)[0]);
#endif

	if (EXPECT_FALSE(a->prec <= GMP_LIMB_BITS))
		goto done_word;

	mploc_z2_init(x,a->f);
	
	/* just for the malloc() */
	mploc_z2_force_precision(x,2*GMP_LIMB_BITS*SIZ(a->x));

	mploc_z2_set_ui(x,w,GMP_LIMB_BITS);
	mploc_z2_inv_improve(x,a);
	mploc_z2_set_mpz(z,x->x,a->prec);
	mploc_z2_clear(x);
	
	ASSERT(mploc_z2_is_good_looking(z));
	
	return;
done_word:
	mploc_z2_set_ui(z,w,a->prec);
	
	ASSERT(mploc_z2_is_good_looking(z));
	
	return;
}

/*
 * mploc_z2_invsqrt -- compute the inverse of the square root
 *
 * We find by Newton iteration a zero of f(x)=1/ax^2-1
 *
 * Iteration is x' = x - (1/ax^2-1)/(-2/ax^3)
 *                 = x + x/2 * (1-ax^2)
 *
 * Given a such that a = 1 mod 8, we return a square root which is equal
 * to 1 mod 4 ; If a is known with precision n bits, x is known with
 * precision n-1 bits.
 *
 * on one round of the newton iteration, the precision goes from n bits
 * to 2n-1 bits. It is easier to start with the easy precision 3 rather
 * than 2 (the result fits better wrt word boundaries, and the init is cheap).
 */

int mploc_z2_invsqrt(mploc_ptr z, mploc_srcptr a)
{
	mploc_t x,y;	/* x is needed in case a == z ; malloc is free anyway */
	mp_limb_t a0;

	ASSERT(mploc_z2_is_good_looking(a));

	if (EXPECT_FALSE(a->prec==0)) {
		mploc_z2_force_precision(z,0);
		return 1;
	}
	
	a0 = PTR(a->x)[0];

	if (EXPECT_FALSE(!(a0&1UL))) {
		DIVIDE_BY_ZERO();
	} else if (EXPECT_FALSE(a->prec<=2)) {
		int r=(a->prec==1) || ((a0 & 3UL) == 1UL);
		if (r) mploc_z2_set(z,a);
		return r;
	} else if (EXPECT_FALSE((a0 & 7UL) != 1UL)) {
		return 0;
	}

	mploc_z2_init(x,a->f);
	mploc_z2_init(y,a->f);

	/* Allocate space here once and for all */
	mploc_z2_force_precision(x,2*a->prec);
	mploc_z2_force_precision(y,2*a->prec);

	mploc_z2_set_ui(x, a0 | (a0>>1), 3);

	for(;x->prec<a->prec-1;) {
		// mploc_z2_print(x,"x");

		mploc_z2_force_precision(x,2*x->prec);
		mploc_z2_sqr(y,x);
		mploc_z2_mul(y,a,y);
		mploc_z2_sub_ui(y,y,1);
		mploc_z2_mul(y,x,y);
		mploc_z2_rshift(y,y,1);
		mploc_z2_sub(x,x,y);	/* Precision loss ! */
	}

	mploc_z2_set_mpz(z,x->x,a->prec-1);

	mploc_z2_clear(y);
	mploc_z2_clear(x);

	ASSERT(mploc_z2_is_good_looking(z));

	return 1;
}

int mploc_z2_sqrt(mploc_ptr z, mploc_srcptr a)
{
	mp_size_t v;
	int r;

	if (mploc_z2_eq_ui(a,0)) {
		mploc_z2_set_change_precision(z,a,MAX(a->prec-1,0));
		return 1;
	}

	v=mploc_z2_valuation(a);

	if (v & 1)
		return 0;

	if (a==z || v) {
		mploc_t tmp;

		mploc_z2_init(tmp,a->f);
		mploc_z2_rshift(tmp,a,v);
		r=mploc_z2_invsqrt(z,tmp);
		if (r) {
			mploc_z2_mul(z,z,tmp);
			mploc_z2_lshift(z,z,v/2);
		}
		mploc_z2_clear(tmp);
	} else {
		r=mploc_z2_invsqrt(z,a);
		if (r) {
			mploc_z2_mul(z,z,a);
		}
	}
	return r;
}

int mploc_z2_eq_ui(mploc_srcptr x, mp_limb_t v)
{
	mploc_t z;
	int res;
	
	ASSERT(mploc_z2_is_good_looking(x));
	
	mploc_z2_init(z,x->f);
	mploc_z2_sub_ui(z,x,v);

	mploc_z2_normalize(z);

	res = mploc_z2_valuation(z)>=z->prec;
	//MPN_NORMALIZE(PTR(z->x),SIZ(z->x));
	//res = mpz_cmp_ui(z->x,0)==0;

	mploc_z2_clear(z);

	return res;
}

int mploc_z2_eq(mploc_srcptr x, mploc_srcptr y)
{
	mploc_t	z;
	int res;
	
	ASSERT(mploc_z2_is_good_looking(x));
	ASSERT(mploc_z2_is_good_looking(y));

	mploc_z2_init(z,x->f);
	mploc_z2_sub(z,x,y);
	
	mploc_z2_normalize(z);
	mploc_z2_insert_zero(z,z->prec);
	MPN_NORMALIZE(PTR(z->x),SIZ(z->x));
	res = mpz_cmp_ui(z->x,0)==0;

	mploc_z2_clear(z);

	return res;
}

/*
 * mploc_z2_clear -- clears an element
 */
void mploc_z2_clear(mploc_ptr x)
{
#ifdef DEBUG_MPLOC
	x->f->people--;
#endif
	mpz_clear(x->x);
	memset(x,0,sizeof(struct mploc_struct));
}

void mploc_z2_trace(mploc_ptr z, mploc_srcptr x)
{
	mploc_z2_set(z,x);
}


