/* $Id: zp.c 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

/* This is a standard mploc<zp> module */

/* -- zp.c --
 *
 * This interface is for silly padics, p!=2, not extended, stored as
 * integers. It creates a lookup table for the precomputed powers of p.
 *
 * The interface for p=2 is really similar, except that there is of
 * course no need for a lookup table of prime powers.
 */

/***************************************************************************/
/* HEADERS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HAVE_MEMSET 1

#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "gmp_hacks.h"
#include "mploc.h"
#include "zp.h"

#include "ff/rf_layer_prime.h"

#include "manu.h"

#if 0
void mpz_mul_wrapper(mpz_ptr z, mpz_srcptr x, mpz_srcptr y)
{
	fprintf(stderr,
		"Doing mul %lu x %lu\n",labs(x->_mp_size),labs(y->_mp_size));
	mpz_mul(z,x,y);
}

#ifdef mpz_mul
#undef mpz_mul
#endif
#define mpz_mul(z,x,y) mpz_mul_wrapper(z,x,y)
#endif


/* The hash table containing the powers of the primes */

/* An entry: */

struct zp_ht_key_s {
	int	l;
	mpz_t	px;
};

typedef struct zp_ht_key_s zp_ht_key[1];

/* The hash table in itself */

struct zp_ht {
	zp_ht_key * keys;
	int	nk;
};

/* This is enough for the private data, specific to the zp module */

struct mplf_private {
	double		logp;	/* An approximation of log_2(p) */
	struct zp_ht	ht[1];
	mploc_t		half;	/* precomputed 1/2 */
};

extern int zp_ht_lookup(struct zp_ht * ht, int l);
extern mpz_ptr zp_power_of_p(struct zp_ht * ht, mpz_ptr p, int l);

/***************************************************************************/
/* HELPER FUNCTIONS */

/* zp_ht_lookup -- Do the lookup.
 *
 * INPUT: The hash table, and an index l to look up
 * OUTPUT: An index x in the hash table such that ht->keys[x-1]<l<=ht->keys[x].
 * (convention : [x-1]=-infty if it does not exist)
 *
 * if x < ht->nk , use l==ht->keys[x] to test whether the key was
 * found
 *
 * This functionality is unfortunately *not* provided as is by bsearch()
 */
int zp_ht_lookup(struct zp_ht * ht, int l)
{
	int a, b, c, h;

	a=0;
	b=ht->nk;

	/* Invariant : ht->keys[a]->l <= l < ht->keys[b]->l 
	 * a and b are within [0..nk]
	 * [b] is never accessed
	 */

	for(;b-a>=2;) {
		c=(a+b)/2;	/* hence a+1<=c<=b-1 */
		h=ht->keys[c]->l;
		if (h==l) {
			return c;
		} else if (h<l) {
			a=c;
		} else { /* l<h */
			b=c;
		}
	}

	return b;
}

mpz_ptr zp_power_of_p(struct zp_ht * ht, mpz_ptr p, int l)
{
	int x;
	x=zp_ht_lookup(ht,l);

	if (x < ht->nk && l==ht->keys[x]->l)
		return ht->keys[x]->px;

	ht->keys=realloc(ht->keys,(ht->nk+1)*sizeof(zp_ht_key));

	if (ht->nk-x) {
		memmove(ht->keys+x+1,ht->keys+x,(ht->nk-x)*sizeof(zp_ht_key));
	}

	ht->keys[x]->l=l;
	mpz_init(ht->keys[x]->px);

	mpz_pow_ui(ht->keys[x]->px,p,l);

	ht->nk++;

	return ht->keys[x]->px;
}
mpz_srcptr p_to_the(int n, mplf_t f)
{
	return zp_power_of_p(f->priv->ht,f->p,n);
}

/***************************************************************************/
/* ENTRY POINTS -- FIELD INTERFACE */

/* 
 * mplf_zp_create -- creates a p-adic integer ring
 *
 * INPUT: The prime p
 * OUTPUT: The struct containing useful information
 */

mplf_t mplf_zp_create(mpz_srcptr p)
{
	struct mplf_struct * f;
	
	f=malloc(sizeof(struct mplf_struct));
	mpz_init_set(f->p,p);
#ifdef DEBUG_MPLOC
	f->people=0;
#endif
	//f->force_precision=&mploc_zp_force_precision;
	f->priv=malloc(sizeof(struct mplf_private));
	mplf_zp_tell_logp(f,(double) mpz_sizeinbase(p,2));
	f->priv->ht->nk=1;
	f->priv->ht->keys=malloc(sizeof(zp_ht_key));
	mpz_init_set(f->priv->ht->keys[0]->px,p);
	f->priv->ht->keys[0]->l=1;
	f->degree=1;
	f->ramified=0;
	ATTACH_RESIDUE_FIELD(f);
	mploc_zp_init(f->priv->half,f);
	return f;
}

/*
 * mplf_zp_destroy -- clears any data associated with a field.
 *
 * INPUT: The field info
 * OUTPUT: none
 * SIDE EFFECT: everything cleaned up ; *pf is set to NULL
 */

void mplf_zp_destroy(mplf_t * pf)
{
	int i;

	mploc_zp_clear((*pf)->priv->half);

#ifdef DEBUG_MPLOC
	ASSERT(!(*pf)->people);
#endif
	for(i=0;i<(*pf)->priv->ht->nk;i++) {
		mpz_clear((*pf)->priv->ht->keys[i]->px);
	}
	DETACH_RESIDUE_FIELD(*pf);
	free((*pf)->priv->ht->keys);
	free((*pf)->priv);
	mpz_clear((*pf)->p);
	free((*pf));
	(*pf)=NULL;
}

/*
 * mplf_zp_tell_logp -- Change the log(p) information
 *
 * INPUT: the field info, and log_2(p) as a double.
 *
 * COMMENTS: Use this when log_2(p) is well away from an integer, in
 * order to save memory.
 *
 */

void mplf_zp_tell_logp(mplf_t f, double x)
{
	f->priv->logp=x;
}

/***************************************************************************/
/* ENTRY POINTS -- ELEMENT INTERFACE */

/*
 * mploc_zp_force_precision -- change the significance of a padic int.
 *
 * INPUT: the padic number and the new precision
 * OUTPUT: none
 * SIDE EFFECT: x->prec changed. Inner representation updated.
 *
 * COMMENTS: - Depending on the prime used, there might be a small memory
 * waste here since the logp field in private stores only a gross
 * approximation of log_2(p). Use mplf_zp_tell_logp to improve this.
 *           - A downgrade in the precision costs a modular reduction.
 */

int mploc_zp_is_good_looking(mploc_srcptr x)
{
	if (SIZ(x->x)<0) {
		fprintf(stderr, "mploc_t represented by negative integer\n");
		return 0;
	}

	if (ALLOC(x->x) < SIZ(x->x)) {
		fprintf(stderr, "mploc_t with insufficient allocation\n");
		return 0;
	}

#if 0	/* Enable only to diagnose missing reduction problems */
	if (ALLOC(x->x) > MAX(4*SIZ(x->x),SIZ(x->x)+65536)) {
		fprintf(stderr, "mploc_t with unexpectedly high allocation\n");
		return 0;
	}
#endif
	
	return 1;
}

static void realloc_on_mpz_t_below(mploc_ptr x)
{
	mp_size_t bits,req;
	bits = x->prec*x->f->priv->logp;
	req=iceildiv(bits,GMP_LIMB_BITS);
	if (ALLOC(x->x) < req) {
		mpz_realloc2(x->x,bits);
		/*
		PTR(x->x)=realloc(PTR(x->x),req*sizeof(mp_limb_t));
		ALLOC(x->x)=req;
		*/
		/* 
		memset(PTR(x->x)+abs(SIZ(x->x)),
				0,abs(req-SIZ(x->x))*sizeof(mp_limb_t));
				*/
	}
}

void mploc_zp_force_precision(mploc_ptr x, mp_size_t k)
{
	/*
	if (x->prec == k) { return; }

	if (x->prec < k) {
		x->prec=k;
		realloc_on_mpz_t_below(x);
	} else {
		x->prec=k;
	}
	*/

	x->prec=k;

	/* We used to normalize here ; now we don't do it
	 * anymore.
	 * XXX Check if this causes bugs.
	 */

	ASSERT(mploc_zp_is_good_looking(x));
}

/*
 * mploc_init -- create an element
 */
void mploc_zp_init(mploc_ptr x, mplf_t f)
{
	mpz_init(x->x);
	x->prec=0;
	x->f=f;
#ifdef DEBUG_MPLOC
	x->f->people++;
#endif
	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_set_change_precision(mploc_ptr x, mploc_srcptr y, mp_size_t k)
{
	ASSERT(mploc_zp_is_good_looking(y));
	x->prec=k;
	mpz_mod(x->x,y->x,p_to_the(k,y->f));
	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_set(mploc_ptr x, mploc_srcptr y)
{
	ASSERT(mploc_zp_is_good_looking(y));
	x->prec=y->prec;
	mpz_set(x->x,y->x);
	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_set_ui(mploc_ptr x, mp_limb_t u, unsigned int prec)
{
	mpz_set_ui(x->x,u);
	x->prec=prec;
	realloc_on_mpz_t_below(x);
	ASSERT(mploc_zp_is_good_looking(x));
}


void mploc_zp_set_mpz(mploc_ptr x, mpz_srcptr a, unsigned int prec)
{
	mpz_set(x->x,a);
	x->prec=prec;
	realloc_on_mpz_t_below(x);
	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_set_random(mploc_ptr x, gmp_randstate_t prng, unsigned int prec)
{
	mpz_urandomm(x->x,prng,p_to_the(prec,x->f));
	x->prec=prec;
	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_add(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	/* The result is computed up to the minimum precision of the
	 * operands. The more precise operand is reduced to the low
	 * precision
	 */

	mpz_ptr	px;
	mpz_t tmp;

	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(y));

	ASSERT(z->f == x->f && x->f && y->f);

	z->prec = MIN(x->prec,y->prec);
	px = zp_power_of_p(y->f->priv->ht, y->f->p, z->prec);

	if (x->prec < y->prec) {
		mpz_init(tmp);
		mpz_mod(tmp,y->x,px);
		mpz_add(z->x,tmp,x->x);
		mpz_clear(tmp);
	} else if (y->prec < x->prec) {
		mpz_init(tmp);
		mpz_mod(tmp,x->x,px);
		mpz_add(z->x,tmp,y->x);
		mpz_clear(tmp);
	} else {
		mpz_add(z->x,x->x,y->x);
	}
	if (mpz_cmp(z->x,px)>=0) {
		mpz_sub(z->x,z->x,px);
	}

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_normalize(mploc_ptr x)
{
	mpz_ptr px;

	px = zp_power_of_p(x->f->priv->ht, x->f->p, x->prec);
	mpz_mod(x->x,x->x,px);

	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_neg(mploc_ptr z, mploc_srcptr x)
{
	ASSERT(mploc_zp_is_good_looking(x));

	mpz_neg(z->x,x->x);
	z->prec=x->prec;
	mploc_zp_normalize(z);

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_sub(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	mploc_t s;
	ASSERT(mploc_zp_is_good_looking(x));

	mploc_zp_init(s,x->f);
	mploc_zp_neg(s,y);
	mploc_zp_add(z,x,s);
	mploc_zp_clear(s);

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_add_ui(mploc_ptr z, mploc_srcptr x, mp_limb_t u)
{
	ASSERT(mploc_zp_is_good_looking(x));

	mpz_add_ui(z->x,x->x,u);
	z->prec=x->prec;

	ASSERT(mploc_zp_is_good_looking(z));
}
void mploc_zp_sub_ui(mploc_ptr z, mploc_srcptr x, mp_limb_t u)
{
	ASSERT(mploc_zp_is_good_looking(x));
	
	mpz_sub_ui(z->x,x->x,u);
	z->prec=x->prec;

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_lshift(mploc_ptr z, mploc_srcptr x, mp_size_t k)
{
	ASSERT(mploc_zp_is_good_looking(x));
	
	if (EXPECT_FALSE(k==0)) {
		mploc_zp_set(z,x);
	}

	z->prec = x->prec + k;
	mpz_mul(z->x,x->x,p_to_the(k,x->f));

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_rshift(mploc_ptr z, mploc_srcptr x, mp_size_t k)
{
	ASSERT(mploc_zp_is_good_looking(x));
	
	if (EXPECT_FALSE(k==0)) {
		mploc_zp_set(z,x);
	}

	z->prec = MAX(x->prec - k,0);
	mpz_fdiv_q(z->x,x->x,p_to_the(k,x->f));

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_sqr(mploc_ptr z, mploc_srcptr x)
{
	mploc_zp_mul(z,x,x);
}

void mploc_zp_mul(mploc_ptr z, mploc_srcptr x, mploc_srcptr y)
{
	ASSERT(mploc_zp_is_good_looking(x));

	BAD_CODE_WARNING;

	mpz_mul(z->x,x->x,y->x);
	mploc_zp_force_precision(z,MIN(x->prec,y->prec));

	// XXX Used to be done in force_precision ; This is not anymore
	// true...
	mploc_zp_normalize(z);

	ASSERT(mploc_zp_is_good_looking(z));
}

void mploc_zp_print(mploc_srcptr x)
{
	ASSERT(mploc_zp_is_good_looking(x));

	if (mpz_cmp_ui(x->f->p,10)<0) {
		printf("In base %lu: ", mpz_get_ui(x->f->p));
		mpz_out_str (stdout, mpz_get_ui(x->f->p), x->x);
	} else {
		mpz_t q,r;
		int i;
		mpz_init(q);
		mpz_init(r);
		mpz_set(q,x->x);
		for(i=0;i<x->prec && mpz_cmp_ui(q,0)!=0;i++) {
			mpz_fdiv_qr(q,r,q,x->f->p);
			gmp_printf("%Zd,",r);
		}
		if (i<x->prec)
			printf("0<%ld>",x->prec-i);
		printf("...");
		mpz_clear(q);
		mpz_clear(r);
	}
	printf(" + O(p^%d)", (int) x->prec);
}

/*
 * mploc_zp_valuation -- compute the valuation
 *
 * This function is expected to return 0 in most cases. Therefore it is
 * not meant to be efficient for the cases where an element has a big
 * valuation.
 *
 * If x->prec is returned, of course one shall assume that the result is
 * not significant (it can be anything between x->prec and +infty)
 */
int mploc_zp_valuation(mploc_srcptr x)
{
	mpz_t q,r;
	int	i;

	ASSERT(mploc_zp_is_good_looking(x));

	mpz_init_set(q,x->x);
	mpz_init_set_ui(r,0);

	for(i=0;i<x->prec;i++) {
		mpz_fdiv_qr(q,r,q,x->f->p);
		if (mpz_cmp_ui(r,0)!=0)
			break;
	}

	mpz_clear(q);
	mpz_clear(r);

	return i;
}

/*
 * mploc_zp_inv -- compute an inverse.
 *
 * We find by Newton iteration a zero of f(x)=ax-1
 *
 * iteration is x <- x - x(a*x-1)
 *
 * There must be a relatively high threshold below which simple modular
 * inversion performs much better.
 *
 * The code in here can be largely improved
 */


static int inv_seed=1;
static int inv_method=2;

void mploc_zp_inversion_configure(int new_method, int new_seed)
{
	inv_method=new_method;
	inv_seed=new_seed;
}

void mploc_zp_inv_improve_iterative(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t	t;

	ASSERT(mploc_zp_is_good_looking(a));
	ASSERT(mploc_zp_is_good_looking(x));
	
	mploc_zp_init(t,a->f);
	mploc_zp_init(y,a->f);

	for(;x->prec<aim;) {

		/* Do x_{n+1} <- x_n - x_n*(a*x_n-1) */
		// mploc_zp_print(x,"x");
		
		/* Within this loop, we have to keep track of the
		 * significance *and* the size of our data. Given the
		 * fact that modular reductions cost a lot, these mustn't
		 * be performed too early: once in the loop is good ;
		 * less in extremely bad (exponential growth !) ; more is
		 * probably not worth it.
		 */
		/* Assume x has precision n, and size N here */
		mpz_mod(t->x,a->x,p_to_the(2*x->prec,a->f));	/* t: 2n,2N */
		
		mpz_mul(y->x,t->x,x->x);		/* y: prec=2n, sz=2N*/
		/* This operation is done *before* increasing x's size ;
		 * therefore the resulting y has prec=2n, size=3N */
		/* All in all, this gives a 2N x N -> 2N op (3N obtained)
		 */

		mpz_sub_ui(y->x,y->x,1);
		
		/* Now we know that p^n divides y ; Alas, we can't really
		 * use it here to speed up things. Shifting p^n out and
		 * in again would be a real pain.
		 *
		 * y has size 3N here, and precision 2n only. Therefore
		 * we could chop off the high part with a mod.
		 * Unfortunately, this doesn't pay off.
		 */
		mpz_mul(y->x,y->x,x->x);
		/* We waste time here, somehow. Now we have
		 * y:pr=2n,sz=4N : that's way too much. */

		/* An alternative path to finish the loop:
		 * mpz_sub(x->x,x->x,y->x);
		 * x->prec = 2*x->prec;
		 * mploc_zp_normalize(x);
		 */

		/* Do a reduction, then ; it's late enough (!)*/
		/* Note that if we had done a reduction before, then one
		 * would be advisable here anyway, since y has at least
		 * sz=3N after the preceding mult */
		mpz_neg(y->x,y->x);	/* almost a no-op */
		mpz_mod(t->x,y->x,p_to_the(2*x->prec,a->f));
		/* t : pr=2n, sz=2N (and valuation=n) */

		mpz_add(x->x,x->x,t->x);
		/* The resulting x is decently normalized ; there is no
		 * need to go further on that path. */
		x->prec*=2;
	}

	mploc_zp_clear(t);
	mploc_zp_clear(y);

	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_inv_improve_recursive(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t	t;

	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(a));

	if (x->prec >= aim) return;

	mploc_zp_init(t,a->f);
	mploc_zp_init(y,a->f);

	mploc_zp_inv_improve_recursive(x,a,aim - aim/2);
	
	mpz_mod(t->x,a->x,p_to_the(aim,a->f));
	mpz_mul(y->x,t->x,x->x);
	mpz_sub_ui(y->x,y->x,1);
	mpz_mul(y->x,y->x,x->x);
	mpz_neg(y->x,y->x);
	mpz_mod(t->x,y->x,p_to_the(aim,a->f));
	mpz_add(x->x,x->x,t->x);
	x->prec = aim;

	mploc_zp_clear(t);
	mploc_zp_clear(y);

	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_inv_improve_recursive2(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t	t;

	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(a));

	if (x->prec >= aim) return;

	mploc_zp_init(t,a->f);
	mploc_zp_init(y,a->f);

	mploc_zp_inv_improve_recursive2(x,a,aim - aim/2);

	mpz_mod(t->x,a->x,p_to_the(aim,a->f));
	mpz_mul(y->x,x->x,x->x);
	mpz_mul(y->x,t->x,y->x);
	mpz_neg(y->x,y->x);
	mpz_mod(t->x,y->x,p_to_the(aim,a->f));
	mpz_mul_2exp(x->x,x->x,1);
	mpz_add(x->x,x->x,t->x);
	x->prec = aim;

	mploc_zp_clear(t);
	mploc_zp_clear(y);

	ASSERT(mploc_zp_is_good_looking(x));
}

#include "../bench/timings.h"

void mploc_zp_inv_improve_recursive3(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t	t;

	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(a));

	if (x->prec >= aim) return;

	mploc_zp_init(t,a->f);
	mpz_mod(t->x,a->x,p_to_the(aim,a->f));
	mploc_zp_inv_improve_recursive3(x,t,aim - aim/2);
	mploc_zp_init(y,a->f);
	mpz_mul(y->x,x->x,x->x);
	mpz_mul(y->x,t->x,y->x);
	mpz_neg(y->x,y->x);
	mpz_mod(t->x,y->x,p_to_the(aim,a->f));
	mpz_mul_2exp(x->x,x->x,1);
	mpz_add(x->x,x->x,t->x);
	x->prec = aim;

	mploc_zp_clear(t);
	mploc_zp_clear(y);

	ASSERT(mploc_zp_is_good_looking(x));
}

static unsigned long xdm1a,xdm1b,xdr,xds,xdmula,xdmulb,xdm2a,xdm2b;

static fasttimer_t xtm1,xtr,xts,xtmul,xtm2,xtot,xboot;

void inv_stats_grab(fasttimer_t res[],unsigned long degrees[])
{
	res[0]=xtm1;
	res[1]=xtr;
	res[2]=xts;
	res[3]=xtmul;
	res[4]=xtm2;
	res[5]=xtot;
	res[6]=xboot;
	xtm1=0;
	xtr=0;
	xts=0;
	xtmul=0;
	xtm2=0;
	xtot=0;
	xboot=0;

	degrees[0]=xdm1a;
	degrees[1]=xdm1b;
	degrees[2]=xdr;
	degrees[3]=xds;
	degrees[4]=xdmula;
	degrees[5]=xdmulb;
	degrees[6]=xdm2b;
	degrees[7]=xdm2b;
	xdm1a=0;
	xdm1b=0;
	xdr=0;
	xds=0;
	xdmula=0;
	xdmulb=0;
	xdm2b=0;
	xdm2b=0; 	
}


void mploc_zp_inv_improve_recursive4(mploc_ptr x, mploc_srcptr a, mp_size_t aim)
{
	mploc_t y;
	mploc_t	t;
	mpz_srcptr p;

	fasttimer_t tm1,tr,ts,tmul,tm2;

	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(a));

	if (x->prec >= aim) return;

	push_time();

	mploc_zp_init(t,a->f);

	p=p_to_the(aim,a->f);
	xdm1a+=SIZ(a->x);
	xdm1b+=SIZ(p);
	mpz_mod(t->x,a->x,p);

	tm1 = lookup_time();

	xdr += SIZ(t->x);
	mploc_zp_inv_improve_recursive3(x,t,aim - aim/2);

	tr = lookup_time();

	mploc_zp_init(y,a->f);
	xds+=SIZ(x->x);
	mpz_mul(y->x,x->x,x->x);

	ts = lookup_time();

	xdmula+=SIZ(t->x);
	xdmulb+=SIZ(y->x);
	mpz_mul(y->x,t->x,y->x);

	tmul = lookup_time();

	mpz_neg(y->x,y->x);

	xdm2a+=SIZ(y->x);
	xdm2b+=SIZ(p);
	mpz_mod(t->x,y->x,p);//p_to_the(aim,a->f));

	tm2 = lookup_time();

	mpz_mul_2exp(x->x,x->x,1);
	mpz_add(x->x,x->x,t->x);

	tm2 -= tmul;
	tmul-= ts;
	ts  -= tr;
	tr  -= tm1;
	
	//{
		//static int i=0;
		//++i;
		xtm1+=tm1;
		xtr+=tr;
		xts+=ts;
		xtmul+=tmul;
		xtm2+=tm2;
		xtot+=lookup_time();
		
	//fprintf(stdout,"%d %lu %llu %llu %llu %llu %llu %llu\n", i, aim, tm1,tr,ts,tmul,tm2,lookup_time());
	//}
	discard_timer();

	x->prec = aim;

	mploc_zp_clear(t);
	mploc_zp_clear(y);

	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_inv_improve(mploc_ptr x, mploc_srcptr a)
{
	ASSERT(mploc_zp_is_good_looking(a));
	ASSERT(mploc_zp_is_good_looking(x));

	switch(inv_method) {
		case 0: mploc_zp_inv_improve_iterative(x,a,a->prec); break;
		case 1: mploc_zp_inv_improve_recursive(x,a,a->prec); break;
		case 2: mploc_zp_inv_improve_recursive2(x,a,a->prec); break;
		case 3: mploc_zp_inv_improve_recursive4(x,a,a->prec); break;
	}

	ASSERT(mploc_zp_is_good_looking(x));
}

void mploc_zp_inv(mploc_ptr z, mploc_srcptr a)
{
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	ASSERT(mploc_zp_is_good_looking(a));

	if (mploc_zp_valuation(a)!=0)
		DIVIDE_BY_ZERO();
		
	if (inv_seed == 0) {
		
		mpz_invert(z->x,a->x, p_to_the(a->prec,a->f));
		z->prec=a->prec;

	} else {
		mploc_t x;

		push_time();

		mploc_zp_init(x,a->f);
		mpz_invert(x->x,a->x, p_to_the(inv_seed,a->f));
		x->prec=inv_seed;

		xboot+=lookup_time();
		discard_timer();

		mploc_zp_inv_improve(x,a);

		mploc_zp_set_mpz(z,x->x,a->prec);

		mploc_zp_clear(x);
	}

	ASSERT(mploc_zp_is_good_looking(z));
}

mploc_srcptr	one_half(mplf_t f, int n)
{
	/* prime fields -> We must precompute 1/2 !! */
	if (f->priv->half->prec < n) {
		mploc_t y;
		mploc_zp_init(y,f);
		mploc_zp_set_ui(y,2,n);
		if (f->priv->half->prec)
			mploc_zp_inv_improve_iterative(f->priv->half,y,n);
		else
			mploc_zp_inv(f->priv->half,y);
		mploc_zp_clear(y);
	}
	return f->priv->half;
}

/*
 * mploc_zp_invsqrt -- compute the inverse of the square root
 *
 * We find by Newton iteration a zero of f(x)=1/ax^2-1
 *
 * Iteration is x' = x - (1/ax^2-1)/(-2/ax^3)
 *                 = x + x/2 * (1-ax^2)
 *
 * Therefore no inversion is required.
 */

int mploc_zp_invsqrt0(mploc_ptr z, mploc_srcptr a)
{
	int has_root;
	mploc_t x,y;
	rf_elt r1,r2;
	mploc_srcptr half;

	ASSERT(mploc_zp_is_good_looking(a));

	r1 = new_rfe(a->f);
	r2 = new_rfe(a->f);

	rfe_set_mploc(r1,a,a->f);
	has_root = RF_sqrt(r2,r1,a->f);

	if (!has_root) {
		delete_rfe(r1,a->f);
		delete_rfe(r2,a->f);
		return 0;
	}

	if (RF_is_zero(r2,a->f)) {
		DIVIDE_BY_ZERO();
	}

	RF_inv(r1,r2,a->f);

	mploc_zp_init(x,a->f);
	mploc_zp_init(y,a->f);
	
	mploc_set_rfe(x,r1,a->f);

	half=one_half(a->f,a->prec);

	for(;x->prec<a->prec;) {
		/* Do x_{n+1} <- x_n - x_n*(a*x_n-1) */
		// mploc_zp_print(x,"x");

		mploc_zp_force_precision(x,2*x->prec);

		mploc_zp_normalize(x);
		mploc_zp_mul(y,x,x);
		mploc_zp_mul(y,a,y);
		mploc_zp_sub_ui(y,y,1);
		mploc_zp_mul(y,x,y);
		mploc_zp_mul(y,y,half);
		mploc_zp_sub(x,x,y);
	}

	mploc_zp_set_mpz(z,x->x,a->prec);

	mploc_zp_clear(y);
	mploc_zp_clear(x);

	ASSERT(mploc_zp_is_good_looking(z));

	return 1;
}

int mploc_zp_invsqrt(mploc_ptr z, mploc_srcptr a)
{
	int has_root;
	mploc_t x,y;
	rf_elt r1,r2;
	mploc_srcptr half;

	ASSERT(mploc_zp_is_good_looking(a));

	r1 = new_rfe(a->f);
	r2 = new_rfe(a->f);

	rfe_set_mploc(r1,a,a->f);
	has_root = RF_sqrt(r2,r1,a->f);

	if (!has_root) {
		delete_rfe(r1,a->f);
		delete_rfe(r2,a->f);
		return 0;
	}

	if (RF_is_zero(r2,a->f)) {
		DIVIDE_BY_ZERO();
	}

	RF_inv(r1,r2,a->f);

	mploc_zp_init(x,a->f);
	mploc_zp_init(y,a->f);
	
	mploc_set_rfe(x,r1,a->f);

	half=one_half(a->f,a->prec);

	for(;x->prec<a->prec;) {
		/* Do x_{n+1} <- x_n - x_n*(a*x_n-1) */
		// mploc_zp_print(x,"x");

		mploc_zp_force_precision(x,2*x->prec);

		mploc_zp_normalize(x);
		mploc_zp_mul(y,x,x);
		mploc_zp_mul(y,a,y);
		mploc_zp_sub_ui(y,y,1);
		mploc_zp_mul(y,x,y);
		mploc_zp_mul(y,y,half);
		mploc_zp_sub(x,x,y);
	}

	mploc_zp_set_mpz(z,x->x,a->prec);

	mploc_zp_clear(y);
	mploc_zp_clear(x);

	ASSERT(mploc_zp_is_good_looking(z));

	return 1;
}

int mploc_zp_sqrt(mploc_ptr z, mploc_srcptr a)
{
	mp_size_t v;
	int r;

	if (mploc_zp_eq_ui(a,0)) {
		mploc_zp_set_change_precision(z,a,a->prec);
		return 1;
	}

	v=mploc_zp_valuation(a);

	if (v & 1)
		return 0;

	if (a==z || v) {
		mploc_t tmp;
		mploc_zp_init(tmp,a->f);
		if (v) mploc_zp_rshift(tmp,a,v);
		r=mploc_zp_invsqrt(z,tmp);
		if (r) {
			mploc_zp_mul(z,z,tmp);
			if (v) mploc_zp_lshift(z,z,v/2);
		}
		mploc_zp_clear(tmp);
	} else {
		r=mploc_zp_invsqrt(z,a);
		if (r) {
			mploc_zp_mul(z,z,a);
		}
	}
	return r;
}

int mploc_zp_eq_ui(mploc_srcptr x, mp_limb_t v)
{
	mploc_t z;
	int res;

	ASSERT(mploc_zp_is_good_looking(x));

	mploc_zp_init(z,x->f);
	mploc_zp_sub_ui(z,x,v);

	mploc_zp_normalize(z);
	res = mpz_cmp_ui(z->x,0)==0;

	mploc_zp_clear(z);

	return res;
}

int mploc_zp_eq(mploc_srcptr x, mploc_srcptr y)
{
	mploc_t	z;
	int res;
	
	ASSERT(mploc_zp_is_good_looking(x));
	ASSERT(mploc_zp_is_good_looking(y));

	mploc_zp_init(z,x->f);
	mploc_zp_sub(z,x,y);
	
	mploc_zp_normalize(z);
	res = mpz_cmp_ui(z->x,0)==0;

	mploc_zp_clear(z);

	return res;
}

/*
 * mploc_zp_clear -- clears an element
 */
void mploc_zp_clear(mploc_ptr x)
{
#ifdef DEBUG_MPLOC
	x->f->people--;
#endif
	mpz_clear(x->x);
	memset(x,0,sizeof(struct mploc_struct));
}
